from pathlib import Path
from signal import Signals
import subprocess

import toml

from . import builder
from . import github
from . import pypi
from .filesystem import try_delete
from .log import logger


DOWNLOAD_SOURCES = {
    'gh': github,
    'github': github,
    'pypi': pypi.Downloader('pypi'),
    'testpypi': pypi.Downloader('testpypi'),
    'pypi-test': pypi.Downloader('testpypi'),  # for backwards compatibility
}


def aliases():
    pyproject = toml.loads(Path('pyproject.toml').read_text())
    return pyproject.get('tool', {}).get('bork', {}).get('aliases', {})


def build():
    builder.dist()
    builder.zipapp()


def clean():
    try_delete('./build')
    try_delete('./dist')
    for name in Path.cwd().glob('*.egg-info'):
        if name.is_dir():
            try_delete(name)


def download(package, release_tag, file_pattern, directory):
    if file_pattern is None or len(file_pattern) == 0:
        raise ValueError('file_pattern must be non-empty.')

    if ':' not in package:
        raise ValueError('Invalid package/repository -- no source given.')

    source, package = package.split(':')

    if source not in DOWNLOAD_SOURCES.keys():
        raise ValueError('Invalid package/repository -- unknown source given.')

    source = DOWNLOAD_SOURCES[source]
    source.download(package, release_tag, file_pattern, directory)


def release(repository_name, dry_run):
    pyproject = toml.load('pyproject.toml')

    try:
        release_dict = pyproject['tool']['bork']['release']
        strip_zipapp_version = release_dict.get('strip_zipapp_version', False)
    except KeyError:
        # Not an error.
        strip_zipapp_version = False

    try:
        release_dict = pyproject['tool']['bork']['release']
        release_to_github = release_dict.get('github', False)
        release_to_pypi = release_dict.get('pypi', True)
    except KeyError:
        release_to_github = False
        release_to_pypi = True

    if not release_to_github and not release_to_pypi:
        print('Configured to release to neither PyPi nor GitHub?')

    if release_to_pypi:
        pypi.upload(repository_name, './dist/*.tar.gz', './dist/*.whl',
                    dry_run=dry_run)

    if release_to_github:
        github.upload('./dist/*.pyz',
                      dry_run=dry_run,
                      strip_zipapp_version=strip_zipapp_version)


def run(alias):
    pyproject = toml.load('pyproject.toml')

    try:
        command = pyproject['tool']['bork']['aliases'][alias]
    except KeyError:
        raise RuntimeError("No such alias: '{}'".format(alias))

    logger().info("Running '%s'", command)

    try:
        subprocess.run(command, check=True, shell=True)

    except subprocess.CalledProcessError as error:
        if error.returncode < 0:
            signal = Signals(- error.returncode)
            msg = "command '{}' exited due to signal {} ({})".format(
                error.cmd, signal.name, signal.value,
            )

        else:
            msg = "bork: command '{}' exited with error code {}".format(
                error.cmd, error.returncode,
            )

        raise RuntimeError(msg) from error



from pathlib import Path
from urllib.request import urlopen

from .log import logger


def download_assets(asset_list, directory, name_key=None, url_key=None):
    if name_key is None:
        name_key = 'name'

    if url_key is None:
        url_key = 'url'

    directory = Path(directory)
    directory.mkdir(parents=True, exist_ok=True)

    log = logger()
    for asset in asset_list:
        name = asset[name_key]
        url = asset[url_key]
        path = directory / name

        contents = urlopen(url).read()

        path.write_bytes(contents)
        log.info("Downloaded '%s'", path)



import configparser
from pathlib import Path
import shutil


def load_setup_cfg():
    setup_cfg = configparser.ConfigParser()
    setup_cfg.read('setup.cfg')
    return setup_cfg


def find_files(globs):
    files = []

    for glob in globs:
        matches = map(str, Path().glob(glob))
        files += list(matches)

    return files


def try_delete(path):
    if Path(path).is_dir():
        shutil.rmtree(path)
    elif Path(path).exists():
        raise RuntimeError("'{}' is not a directory".format(path))



import json
from pathlib import Path
import subprocess
import urllib

from .log import logger


class GithubApi:
    def __init__(self, owner, repo, token):
        self.owner = owner
        self.repo = repo
        self.token = token

    def create_release_with_file(self, tag_name, commitish, local_file,
                                 remote_file):
        response1 = self.create_release(tag_name, commitish)
        upload_url = json.loads(response1)['upload_url']
        # Is there a better way to do this `upload_url.replace(...)`?
        upload_url = upload_url.replace('{?name,label}', '')
        response2 = self.add_release_asset(upload_url, local_file, remote_file)
        return response2

    def create_release(self, tag_name, commitish, body=None, prerelease=False):
        if commitish is None:
            commitish = self.run('git', 'rev-parse', 'HEAD')

        if body is None:
            body = tag_name + ' release.'

        logger().info('Creating GitHub release %s. (commit=%s)', tag_name, commitish)

        request = {
            'tag_name': tag_name,
            'target_commitish': commitish,
            'name': tag_name,
            'body': body,
            'draft': False,
            'prerelease': prerelease,
        }
        url = '/repos/{}/{}/releases'.format(self.owner, self.repo)
        response = self._api_post(url, request)
        return response

    def add_release_asset(self, upload_url, local_file, name):
        logger().info('Adding asset %s to release (original file: %s).',
                      name, local_file)

        data = Path(local_file).read_bytes()

        headers = {
            'Content-Type': 'application/octet-stream',
        }

        url = '{}?name={}'.format(upload_url, name)
        response = self._api_post(url, data, headers=headers, server='')
        return response

    @staticmethod
    def run(*command):
        return subprocess.check_output(command).decode().strip()

    # pylint: disable=too-many-arguments

    def _api_post(self, endpoint, data, headers=None, server=None, method=None):
        if method is None:
            method = 'POST'
        if headers is None:
            headers = {}
        if server is None:
            server = 'https://api.github.com'

        headers['Authorization'] = 'token ' + self.token
        headers['Accept'] = 'application/vnd.github.v3+json'

        if isinstance(data, (dict, list)):
            data = json.dumps(data).encode()

        req = urllib.request.Request(server + endpoint, data=data,
                                     headers=headers, method=method)
        logger().debug('%s %s', req.method, req.full_url)

        response = urllib.request.urlopen(req).read().decode()
        return response

    # pylint: enable=too-many-arguments

    def _api_get(self, endpoint, headers=None, server=None):
        return self._api_post(endpoint, None, headers, server, 'GET')



from pathlib import Path
import shutil
import subprocess
import sys
# Slight kludge so we can have a function named zipapp().
import zipapp as Zipapp  # noqa: N812

import pep517.build  # type: ignore
import toml

from .filesystem import load_setup_cfg, try_delete


# The "proper" way to handle the default would be to check python_requires
# in setup.cfg. But, since Bork needs Python 3, there's no point.
DEFAULT_PYTHON_INTERPRETER = '/usr/bin/env python3'


def dist():
    """Build the sdist and wheel distributions."""

    # We can use the pep517 library to add separate source and binary build
    # functions, but I'm leaving that alone unless someone asks for it.
    args = ['--source', '--binary']
    pep517.build.main(pep517.build.parser.parse_args(['.', *args]))


def version_from_sdist_file():
    sdist = max(Path.cwd().glob('dist/*.tar.gz')).name
    return sdist.replace('.tar.gz', '').split('-')[1]


def _prepare_zipapp_directory(source, dest, name):
    ignore = shutil.ignore_patterns('**/__pycache__/**')
    try_delete(dest)
    Path(dest).mkdir()
    realdest = Path(dest, name)
    shutil.copytree(source, realdest, ignore=ignore)

    return Path(realdest).is_dir()


def _zipapp_add_deps(dest):
    config = load_setup_cfg()
    deps = None
    if 'options' in config:
        options = config['options']
        if 'install_requires' in options:
            deps = options['install_requires'].strip().split('\n')
            deps = list(map(str.strip, deps))

    if not deps:
        return

    cmd = [sys.executable, '-m', 'pip', 'install', '--target', dest] + deps
    subprocess.check_call(cmd)


# BAD ASSUMPTION: We assume dist() is called before zipapp().
def zipapp():
    """Build a zipapp for the project."""

    pyproject = toml.load('pyproject.toml')
    config = pyproject.get('tool', {}).get('bork', {})
    zipapp_cfg = config.get('zipapp', {})
    want_zipapp = zipapp_cfg.get('enabled', False)

    if not want_zipapp:
        return

    setup_cfg = load_setup_cfg()

    if 'metadata' not in setup_cfg or 'name' not in setup_cfg['metadata']:
        raise RuntimeError(
            "The [metadata] section of setup.cfg needs the 'name' key set.",
        )

    name = setup_cfg['metadata']['name']

    # The code is assumed to be in ./<name>
    orig_source = str(Path(name))
    source = str(Path('build', 'zipapp'))

    version = version_from_sdist_file()

    # Output file is dist/<package name>-<package version>.pyz.
    target = 'dist/{}-{}.pyz'.format(name, version)

    # To override the default interpreter, add this to your project's setup.cfg:
    #
    # [bork]
    # python_interpreter = /path/to/python
    interpreter = config.get('python_interpreter', DEFAULT_PYTHON_INTERPRETER)

    # This is where GitHub issue #9 ("Allow specifying console_script
    # entrypoint") would likely be implemented.
    main = zipapp_cfg['main']

    _prepare_zipapp_directory(orig_source, source, name)
    _zipapp_add_deps(source)

    Zipapp.create_archive(source, target, interpreter, main)
    if not Path(target).exists():
        raise RuntimeError('Failed to build zipapp: {}'.format(target))



from distutils.version import LooseVersion
import fnmatch
import json
from urllib.request import urlopen

from .asset_manager import download_assets
from .filesystem import find_files
# from .github_api import GithubApi
from .log import logger


def upload(*globs, dry_run=False, strip_zipapp_version=False):
    log = logger()

    files = find_files(globs)

    log.info('Uploading files to Github.')
    for filename in files:
        if strip_zipapp_version and filename.endswith('.pyz'):
            dest_filename = '-'.join(filename.split('-')[:-1]) + '.pyz'
            log.info('- %s as %s', filename, dest_filename)
        else:
            dest_filename = filename
            log.info('- %s', filename)

    if dry_run:
        log.warning('Skipping GitHub upload step since this is a dry run.')
        return

    raise NotImplementedError('github.upload() is not yet implemented')


def _relevant_asset(asset, file_pattern):
    file_patterns = file_pattern.split(',')
    for pattern in file_patterns:
        if fnmatch.fnmatch(asset['name'], pattern):
            return True
    return False


def _get_release_info(repo, name, draft=False, prerelease=False):
    if '/' not in repo:
        raise ValueError(
            "repo must be of format <user>/<repo>, got '{}'".format(repo),
        )

    log = logger()
    url = 'https://api.github.com/repos/{}/releases'.format(repo)
    req = urlopen(url).read().decode()
    releases = json.loads(req)

    try:
        if name == 'latest':
            # Filter out prereleases and drafts (unless specified in the arguments)
            releases = (
                r for r in releases
                if (draft or not r['draft'])
                and (prerelease or not r['prerelease'])  # noqa: W503
            )
            # Find the latest
            release = max(
                releases,
                key=lambda x: LooseVersion(x['tag_name'].lstrip('v')),
            )
            log.info("Selected release '%s' as latest", release['name'])
        else:
            release = list(filter(lambda x: x['tag_name'] == name, releases))[0]

    except (IndexError, ValueError) as e:
        raise RuntimeError("No such Github release: '{}'".format(name)) from e

    return release


def download(repo, release, file_pattern, directory):
    release_info = _get_release_info(repo, release, file_pattern)
    assets = filter(lambda x: _relevant_asset(x, file_pattern),
                    release_info['assets'])
    download_assets(assets, directory, url_key='browser_download_url')



import inspect
import logging
import os
import sys

import click
from click import BadParameter

from . import __version__
from . import api
from .log import logger


DOWNLOAD_SOURCES_STR = ' '.join(api.DOWNLOAD_SOURCES)


@click.group()
def cli():
    pass


@cli.command()
def aliases():
    alias_list = api.aliases()
    if len(alias_list.keys()) == 0:
        print('No aliases available.')
        return

    for key in alias_list.keys():
        print(key)


@cli.command()
def build():
    api.build()


@cli.command()
def clean():
    api.clean()


@cli.command()
@click.option('--files', default='*.pyz',
              help='Comma-separated list of filenames to download. Supports '
                   'wildcards (* = everything, ? = any single character).')
@click.option('--directory', default='downloads',
              help='Directory to save files in. Created if missing.')
@click.argument('package', nargs=1)
@click.argument('release_tag', nargs=1, default='latest')
def download(files, directory, package, release_tag):
    # NOTE: We change the order of the arguments here, to move away from
    #       what makes sense on a CLI interface to what makes sense in a
    #       Python interface.
    try:
        api.download(package, release_tag, files, directory)
    except ValueError as error:
        raise BadParameter(error)


@cli.command()
@click.option('--pypi-repository', default='pypi',
              help='Repository to use. Valid values are pypi, testpypi, or'
                   'anything defined in ".pypirc".')
@click.option('--test-pypi', is_flag=True, default=False,
              help='Release to test.pypi.org instead of pypi.org.\n'
                   'Equivalent to `--pypi-repository testpypi`.')
@click.option('--dry-run', is_flag=True, default=False,
              help="Don't actually release, just show what a release would do.")
def release(pypi_repository, test_pypi, dry_run):
    if test_pypi:
        pypi_repository = 'testpypi'
    api.release(pypi_repository, dry_run)


@cli.command()
@click.argument('alias', nargs=1)
def run(alias):
    api.run(alias)


def main():
    verbose = '--verbose' in sys.argv
    if verbose:
        sys.argv.remove('--verbose')

    debug = '--debug' in sys.argv
    if debug:
        sys.argv.remove('--debug')

    if '--version' in sys.argv:
        print('bork v{}'.format(__version__))
        sys.exit()

    try:
        # pylint: disable=import-outside-toplevel
        import coloredlogs  # type: ignore
        # pylint: enable=import-outside-toplevel

        # Default to only printing WARNING and higher severity messages.
        log_level = logging.WARNING

        # If we got '--verbose', print INFO and higher severity messages.
        if verbose:
            log_level = logging.INFO

        # If we got '--debug', print DEBUG and higher severity messages.
        if debug:
            log_level = logging.DEBUG

        coloredlogs.install(
            level=log_level,
            fmt='%(name)s %(levelname)s %(message)s',
        )

    except ModuleNotFoundError:
        logging.basicConfig(
            level=logging.INFO if verbose else logging.WARNING,
            format='%(name)s %(levelname)s %(message)s',
        )

    try:
        cli()
    except RuntimeError as err:
        thrower = inspect.trace()[-1]
        log = logger(thrower)

        (log.exception if verbose else log.error)(str(err))

        sys.exit(1)


def zipapp_main():
    # If Bork is put in a zipapp, this allows scripts executed as subprocesses
    # to access pep517.compat.
    #
    # The problem area is the `import compat` line in pep517's _in_process.py.
    # https://github.com/pypa/pep517/blob/master/pep517/_in_process.py
    os.environ['PYTHONPATH'] = ':'.join([*sys.path, sys.argv[0] + '/pep517'])

    main()



import inspect
import logging
import sys
from typing import Any, Callable, cast, Dict, Optional, TypeVar


# Index mapping file names to module names.
_module_index: Dict[str, str] = {}


def _get_module(context: inspect.FrameInfo) -> str:
    """Find the module name for a given FrameInfo.

    Raises KeyError if the module isn't loaded.

    Pilfered from Jamie Bliss' equivalent in pursuedpybear:
      https://github.com/ppb/pursuedpybear/blob/master/ppb/utils.py
    """
    global _module_index  # pylint: disable=global-statement
    if context.filename not in _module_index:
        _module_index = {
            mod.__file__: mod.__name__
            for mod in sys.modules.values()
            if hasattr(mod, '__file__') and hasattr(mod, '__name__')
        }

    return _module_index[context.filename]


def logger(context: Optional[inspect.FrameInfo] = None) -> logging.Logger:
    """Provide a logger with a name appropriate for a given context.

    The default context is the caller's.
    """
    if context is None:
        context = inspect.getframeinfo(inspect.currentframe().f_back)  # type: ignore

    if context is None:
        # This may not be the best way to handle it, but if we get here
        # it'll avoid causing errors, at least.
        return logging.getLogger('??.??')

    return logging.getLogger('{}.{}'.format(_get_module(context), context.function))


F = TypeVar('F', bound=Callable[..., Any])


def trace(func: F, level: int = logging.DEBUG) -> F:
    """Decorator to log function entry and exit."""
    log = logging.getLogger('{}.{}'.format(func.__module__, func.__qualname__))

    def wrapper(*args, **kwargs):
        arglist = map(repr, args) + [
            '{}={}'.format(repr(k), repr(v)) for k, v in kwargs
        ]

        try:
            log.log(level, 'called with %s', ', '.join(arglist))
            return func(*args, **kwargs)
        finally:
            log.log(level, 'exiting')

    # Cast required, as mypy cannot (yet?) type decorators usefully:
    #  https://github.com/python/mypy/issues/3157
    return cast(F, wrapper)



import configparser
from pathlib import Path

from twine.cli import dispatch as twine_dispatch  # type: ignore

from .asset_manager import download_assets
from .filesystem import find_files
from .log import logger
from .pypi_api import get_download_info


class Downloader:  # pylint: disable=too-few-public-methods
    # In the future, if we decide this should be public functionality,
    # this lets us choose a better key name with minimal impact.
    PYPIRC_KEY = '_bork-download-endpoint'

    PYPIRC_DEFAULTS = {
        'pypi': {
            PYPIRC_KEY: 'https://pypi.org/simple/',
        },
        'testpypi': {
            PYPIRC_KEY: 'https://test.pypi.org/simple/',
        },
    }

    def __init__(self, repository_name):
        self.repository_url = self._download_url(repository_name)

    def download(self, package, release, file_pattern, directory):
        asset_list = get_download_info(self.repository_url, package, release,
                                       file_pattern)
        download_assets(asset_list, directory)

    def _download_url(self, repository_name):
        path = Path('~/.pypirc').expanduser()
        config = configparser.ConfigParser()
        if path.exists():
            config.read(str(path))

        if repository_name in config:
            repo_config = config.get(repository_name, None)
            if self.PYPIRC_KEY in repo_config:
                return repo_config[self.PYPIRC_KEY]

        return self.PYPIRC_DEFAULTS[repository_name][self.PYPIRC_KEY]


def download(repository_name, package, release, file_pattern, directory):
    return Downloader(repository_name).download(package, release, file_pattern,
                                                directory)


def upload(repository_name, *globs, dry_run=False):
    files = find_files(globs)
    logger().info(
        "Uploading files to PyPI instance '%s': %s",
        repository_name,
        ', '.join(("'{}'".format(file) for file in files)),
    )

    if dry_run:
        logger().warning('Skipping PyPI upload step since this is a dry run.')
    else:
        twine_dispatch(['upload', '--repository', repository_name, *files])



import fnmatch
from html.parser import HTMLParser
import re
from typing import Dict
from urllib.request import urlopen


class SimplePypiParser(HTMLParser):
    files: Dict[str, str] = {}
    in_anchor = False
    current_url = None

    def handle_starttag(self, tag, attrs):
        if tag != 'a':
            return

        self.in_anchor = True
        for attr in attrs:
            if attr[0] == 'href':
                self.current_url = attr[1]

    def handle_endtag(self, tag):
        if tag == 'a':
            self.in_anchor = False
            self.current_url = None

    def handle_data(self, data):
        if not self.in_anchor:
            return

        data = data.strip()
        self.files[data] = self.current_url

    def error(self, message):
        raise RuntimeError('{}: {}'.format(self.__class__.__name__, message))


def _normalize(name):
    return re.sub(r'[-_.]+', '-', name).lower()


def _matches_pattern(filename, file_pattern):
    file_patterns = file_pattern.split(',')

    for pattern in file_patterns:
        if fnmatch.fnmatch(filename, pattern):
            return True
    return False


def _matches_version(filename, version):
    # If it's of the form vX.X.X where X is always a number, remove the
    # leading v.
    # This means that for common versioning styles, the exact content
    # of a GitHub tag can be passed in here.
    if re.match(r'^v[\d.]+$', version):
        return _matches_version(filename, version[1:])

    # HACK: Change .tar.gz to .tgz so we can just remove the last .<blah>.
    filename = filename.replace('.tar.gz', '.tgz')
    filename = '.'.join(filename.split('.')[0:-1])

    file_version = filename.split('-')[1]

    return fnmatch.fnmatch(file_version, version)


def get_download_info(base_url, package, release, file_pattern):
    results = []

    # Normalize base_url so it never ends with a forward slash.
    if base_url.endswith('/'):
        base_url = base_url[:-1]

    html = urlopen('{}/{}/'.format(base_url, package)).read().decode()
    parser = SimplePypiParser()
    parser.feed(html)
    files = parser.files

    keys = list(files.keys()).copy()
    for name in keys:
        if not _matches_pattern(name, file_pattern):
            continue

        if not _matches_version(name, release):
            continue

        # If we get here, we have a valid version.
        results.append({'name': name, 'url': files[name]})

    return results
